import sys
stderr = sys.stderr
sys.stderr = open('/dev/null', 'w')
import keras
sys.stderr = stderr 

import tensorflow as tf
from tensorflow.python.client import device_lib

from ergo.version import banner, __version__

def get_pads(devs):
    namepad = 0
    typepad = 0
    for dev in devs:
        lname = len(dev.name)
        ltype = len(dev.device_type)

        if lname > namepad:
            namepad = lname

        if ltype > typepad:
            typepad = ltype

    return namepad, typepad

# https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
def mem_fmt(num, suffix='B'):
    num = int(num)
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Yi', suffix)

def action_info(argc, argv):
    print(banner.strip("\n") % (__version__, keras.__version__, tf.__version__))
    print("")
    print("Hardware:\n")

    devs = device_lib.list_local_devices()
    npad, tpad  = get_pads(devs)
    for dev in devs:
        print( "%s (%s) - %s" % (dev.name.ljust(npad,' '), dev.device_type.ljust(tpad,' '), mem_fmt(dev.memory_limit)))



import os
import argparse
import logging as log

from ergo.project import Project

def probability(x):
    x = float(x)
    if x < 0 or x > 1:
        raise argparse.ArgumentTypeError("%r not in range [0.0 - 1.0]" % x )
    return x

def usage():
    print("usage: ergo train <path> [args]")
    quit()

def validate_args(args):
    if args.dataset is not None and (not args.dataset.startswith('sum://') and not os.path.exists(args.dataset)):
        log.error("file %s does not exist" % args.dataset)
        quit()

    p_train = 1 - args.test - args.validation
    if p_train < 0:
        log.error("validation and test proportion are bigger than 1")
        quit()
    elif p_train < 0.5:
        log.error("using less than 50% of the dataset for training")
        quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(description="trainer")
    parser.add_argument("-d", "--dataset", dest = "dataset", action = "store", type = str, required = False)
    parser.add_argument("-g", "--gpus", dest = "gpus", action = "store", type = int, default = 0)
    parser.add_argument("-t", "--test", dest = "test", action = "store", type = probability, default = 0.15, required = False)
    parser.add_argument("-v", "--validation", dest = "validation", action = "store", type = probability, default = 0.15, required = False)
    parser.add_argument("--no-save", dest="no_save", action="store_true", default = False, help = "Do not save temporary datasets on disk.")
    args = parser.parse_args(argv)
    validate_args(args)
    return args

def action_train(argc, argv):
    if argc < 1:
        usage()

    prj = Project(argv[0])
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    args = parse_args(argv[1:])
    if args.dataset is not None:
        # a dataset was specified, split it and generate
        # the subsets
        prj.dataset.do_save = not args.no_save
        prj.prepare(args.dataset, args.test, args.validation)
    elif prj.dataset.exists():
        # no dataset passed, attempt to use the previously
        # generated subsets
        prj.dataset.load()
    else:
        log.error("no test/train/validation subsets found in %s, please specify a --dataset argument", argv[0])
        quit()

    prj.train(args.gpus)



import os
import sys
import json
import re
import logging as log

import numpy as np
import pandas as pd
from sklearn.metrics import classification_report, confusion_matrix

from keras.models import model_from_yaml, load_model
from keras.utils.vis_utils import plot_model
from keras.utils.training_utils import multi_gpu_model

import sumpy

from ergo.core.logic import Logic
from ergo.core.utils import clean_if_exist

from ergo.dataset import Dataset
from ergo.templates import Templates

class Project(object):
    @staticmethod
    def create(path):
        log.info("initializing project %s ...", path)
        for filename, data in Templates.items():
            log.info( "creating %s", filename)
            with open( os.path.join(path, filename), 'wt' ) as fp:
                fp.write(data.strip())

    @staticmethod
    def clean(path, full):
        Dataset.clean(path)
        if full:
            # clean everything
            clean_if_exist(path, ( \
                '__pycache__',
                'logs',
                'model.yml',
                'model.png',
                'model.h5',
                'model.fdeep',
                'model.stats',
                'history.json'))

    def __init__(self, path):
        # base info
        self.path  = os.path.abspath(path)
        self.logic = Logic(self.path)
        # model related data
        self.model          = None
        self.accu           = None
        self.model_path     = os.path.join(self.path, 'model.yml')
        self.model_img_path = os.path.join(self.path, 'model.png')
        self.weights_path   = os.path.join(self.path, 'model.h5')
        self.fdeep_path     = os.path.join(self.path, 'model.fdeep')
        # training related data
        self.dataset        = Dataset(self.path)
        self.stats_path     = os.path.join(self.path, 'model.stats')
        self.history_path   = os.path.join(self.path, 'history.json')
        self.history        = None
        self.what           = {
            'train' : "Training --------------------------------------------\n",
            'val'   : "Validation ------------------------------------------\n",
            'test'  : "Test ------------------------------------------------\n"
        }

    def exists(self):
        return os.path.exists(self.path)

    def is_trained(self):
        return os.path.exists(self.weights_path)

    def load(self):
        log.info("loading project %s ..." % self.path)

        if not self.exists():
            return "%s does not exist" % self.path

        err = self.logic.load()
        if err is not None:
            return err

        if os.path.exists(self.weights_path):
            log.debug("loading model from %s ...", self.weights_path)
            self.model = load_model(self.weights_path)
            # https://github.com/keras-team/keras/issues/6462
            self.model._make_predict_function()

        elif os.path.exists(self.model_path):
            log.debug("loading model from %s ...", self.model_path)
            with open(self.model_path, 'r') as fp:
                self.model = model_from_yaml(fp.read())

        else:
            self.model = self.logic.builder(True)
                
        if os.path.exists(self.history_path):
            log.debug("loading history from %s ...", self.history_path)
            with open(self.history_path, 'r') as fp:
                self.history = json.loads(fp.read())
        
        return None
    
    def accuracy_for(self, X, Y, repo_as_dict = False):
        Y_tpred = np.argmax(self.model.predict(X), axis = 1)
        repo    = classification_report(np.argmax(Y, axis = 1), Y_tpred, output_dict = repo_as_dict)
        cm      = confusion_matrix(np.argmax(Y, axis = 1), Y_tpred)
        return repo, cm

    def accuracy(self):
        train, tr_cm = self.accuracy_for(self.dataset.X_train, self.dataset.Y_train)
        test,  ts_cm = self.accuracy_for(self.dataset.X_test, self.dataset.Y_test)
        val,  val_cm = self.accuracy_for(self.dataset.X_val, self.dataset.Y_val)
        return {'train': (train, tr_cm), 
                'test': (test, ts_cm), 
                'val': (val, val_cm)}

    def _save_model(self):
        log.info("updating %s ...", self.model_path)
        with open( self.model_path, 'w' ) as fp:
            fp.write(self.model.to_yaml())

        log.info("updating %s ...", self.weights_path)
        self.model.save(self.weights_path)

    def _save_history(self):
        log.info("updating %s ...", self.history_path)
        with open(self.history_path, 'w') as fp:
            json.dump(self.history, fp) 

    def _out_stats(self, where):
        for who, header in self.what.items():
            vals = self.accu[who]
            where.write( header )
            where.write( vals[0] )
            where.write("\n\n")
            where.write("confusion matrix:")
            where.write("\n\n")
            where.write("%s\n" % vals[1])
            where.write("\n")

    def _save_stats(self):
        log.info("updating %s ...", self.stats_path)
        with open(self.stats_path, 'w') as fp:
            self._out_stats(fp)

    def _from_file(self, filename):
        log.info("preparing data from %s ...", filename)
        return self.logic.prepare_dataset(filename)

    def _from_sum(self, source):
        m = re.findall(r"^sum:\/\/([^@]+)@([^:]+:\d+)$", source)
        if m is None:
            raise Exception("no valid source provided, format is: 'sum:///etc/sumd/creds/cert.pem@localhost:50051'")

        conn = m[0][1]
        cert = m[0][0]

        log.info("connecting to sumd instance %s using certificate %s ...", conn, cert)
        
        size = 100 * 1024 * 1024
        opts = [('grpc.max_send_message_length', size), ('grpc.max_receive_message_length', size)]

        cli      = sumpy.Client(conn, cert, opts=opts)
        per_page = 4096
        count    = cli.list_records(0, per_page)
        left     = count.total
        page     = 1
        data     = []

        log.info("fetching %d records (%d pages)...", count.total, count.pages)

        while left > 0:
            log.debug("  page %d/%d (left %d records)", page, count.pages, left)
            resp = cli.list_records(page, per_page)
            page += 1
            for r in resp.records:
                data.append(r.data)
                left -= 1

        return pd.DataFrame(data)

    def prepare(self, source, p_test, p_val):
        if source.startswith('sum://'):
            data = self._from_sum(source)
        else:
            data = self._from_file(source)

        log.info("data shape: %s", data.shape)

        return self.dataset.source(data, p_test, p_val)

    def train(self, gpus):
        # async datasets saver might be running, wait before training
        self.dataset.saver.wait()
        
        # train
        if self.model is None:
            self.model = self.logic.builder(True)

        to_train = self.model
        if gpus > 1:
            log.info("training with %d GPUs", gpus)
            to_train = multi_gpu_model(self.model, gpus=gpus)

        self.history  = self.logic.trainer(to_train, self.dataset).history
        self.accu     = self.accuracy() 

        print("")
        self._out_stats(sys.stdout)

        # save model structure and weights
        self._save_model()
        # save training history
        self._save_history()
        # save model accuracy statistics
        self._save_stats() 

    """
    unused, zeroing the dataset is better and model agnostic

    def null_feature(self, idx):
        input_layer = self.model.layers[0]
        # get_weights returns a copy!
        input_weights = input_layer.get_weights()
        backup_w = input_weights[0][idx,:].copy()
        backup_b = input_weights[1][idx].copy()
        # set to 0 the weights and bias of this input column
        input_weights[0][idx,:] = 0
        input_weights[1][idx] = 0
        input_layer.set_weights(input_weights)
        return backup_w, backup_b

    def restore_feature(self, idx, backup_w, backup_b):
        input_layer = self.model.layers[0]
        input_weights = input_layer.get_weights()
        input_weights[0][idx,:] = backup_w
        input_weights[1][idx] = backup_b
        input_layer.set_weights(input_weights)
    """

    def view(self):
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg

        if self.model is not None:
            self.model.summary()
            log.info("saving model to %s ...", self.model_img_path)
            plot_model( self.model, 
                    to_file = self.model_img_path,
                    show_shapes = True, 
                    show_layer_names = True)
            img = mpimg.imread(self.model_img_path)
            plt.figure()
            plt.imshow(img)

        if self.history is not None:
            plt.figure()
            # Plot training & validation accuracy values
            plt.subplot(2,1,1)
            plt.plot(self.history['acc'])
            plt.plot(self.history['val_acc'])
            plt.title('Model accuracy')
            plt.ylabel('Accuracy')
            plt.xlabel('Epoch')
            plt.legend(['Train', 'Test'], loc='lower right')
            
            # Plot training & validation loss values
            plt.subplot(2,1,2)
            plt.plot(self.history['loss'])
            plt.plot(self.history['val_loss'])
            plt.title('Model loss')
            plt.ylabel('Loss')
            plt.xlabel('Epoch')
            plt.legend(['Train', 'Test'], loc='upper right')

            plt.tight_layout()
        
        plt.show()






import argparse
import logging as log
import time
import numpy as np

from terminaltables import AsciiTable

from ergo.project import Project

def usage():
    print("usage: ergo relevance <path> --dataset <path> --attributes <path>")
    quit()

def validate_args(args):
    if args.ratio > 1.0 or args.ratio <= 0:
        log.error("ratio must be in the (0.0, 1.0] interval")
        quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(description="relevance")
    parser.add_argument("-d", "--dataset", dest = "dataset", action = "store", type = str, required = True)
    parser.add_argument("-a", "--attributes", dest = "attributes", action = "store", type = str, required = False)
    parser.add_argument("-r", "--ratio", dest = "ratio", action = "store", type = float, required = False, default = 1.0)
    args = parser.parse_args(argv)
    validate_args(args)
    return args

def get_attributes(filename, ncols):
    attributes = []
    if filename is not None:
        with open(filename) as f:
            attributes = f.readlines()
        attributes = [name.strip() for name in attributes]
    else:
        attributes = ["feature %d" % i for i in range(0, ncols)]

    return attributes

def zeroize_feature(X, col, is_scalar_input):
    if is_scalar_input:
        backup = X[:, col].copy()
        X[:,col] = 0
    else:
        backup = X[col].copy()
        X[col] = np.zeros_like(backup)

    return backup

def restore_feature(X, col, backup, is_scalar_input):
    if is_scalar_input:
        X[:,col] = backup
    else:
        X[col] = backup

def action_relevance(argc, argv):
    if argc < 3:
        usage()

    args = parse_args(argv[1:])
    prj = Project(argv[0])
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this project")
        quit()

    prj.prepare(args.dataset, 0.0, 0.0)

    X, y         = prj.dataset.subsample(args.ratio)
    nrows, ncols = X.shape if prj.dataset.is_flat else (X[0].shape[0], len(X))
    attributes   = get_attributes(args.attributes, ncols)

    log.info("computing relevance of %d attributes on %d samples ...", ncols, nrows)

    start = time.time()

    ref_accu, ref_cm = prj.accuracy_for(X, y, repo_as_dict = True)
    deltas           = []
    tot              = 0
    speed            = (1.0 / (time.time() - start)) * nrows

    for col in range(0, ncols):
        log.info("[%.2f evals/s] computing relevance for attribute [%d/%d] %s ...", speed, col + 1, ncols, attributes[col])

        backup = zeroize_feature(X, col, prj.dataset.is_flat)

        start = time.time()

        accu, cm = prj.accuracy_for(X, y, repo_as_dict = True)

        speed = (1.0 / (time.time() - start)) * nrows

        delta = ref_accu['weighted avg']['precision'] - accu['weighted avg']['precision']
        tot  += delta

        deltas.append((col, delta))

        restore_feature(X, col, backup, prj.dataset.is_flat)

    deltas = sorted(deltas, key = lambda x: abs(x[1]), reverse = True)

    num_irrelevant = 0
    table = [("Column", "Feature", "Relevance")]
    for delta in deltas:
        col, d = delta
        if d != 0.0:
            relevance = (d / tot) * 100.0
            row       = ("%d" % col, attributes[col], "%.2f%%" % relevance)
            row       = ["\033[31m%s\033[0m" % e for e in row] if relevance < 0.0 else row
            table.append(row)
        else:
            num_irrelevant += 1

    print("")
    print(AsciiTable(table).table)
    print("")

    if num_irrelevant > 0:
        log.info("%d features have 0 relevance.", num_irrelevant)



#!/usr/bin/env python3
"""Convert a Keras model to frugally-deep format.
"""

import base64
import datetime
import hashlib
import json
import sys

import keras
import numpy as np
from keras import backend as K
from keras.layers import Input
from keras.models import Model, load_model

import logging as log
from ergo.project import Project

__author__ = "Tobias Hermann"
__copyright__ = "Copyright 2017, Tobias Hermann"
__license__ = "MIT"
__maintainer__ = "Tobias Hermann, https://github.com/Dobiasd/frugally-deep"
__email__ = "editgym@gmail.com"

STORE_FLOATS_HUMAN_READABLE = False


def transform_input_kernel(kernel):
    """Transforms weights of a single CuDNN input kernel into the regular Keras format."""
    return kernel.T.reshape(kernel.shape, order='F')


def transform_recurrent_kernel(kernel):
    """Transforms weights of a single CuDNN recurrent kernel into the regular Keras format."""
    return kernel.T


def transform_kernels(kernels, n_gates, transform_func):
    """
    Transforms CuDNN kernel matrices (either LSTM or GRU) into the regular Keras format.

    Parameters
    ----------
    kernels : numpy.ndarray
        Composite matrix of input or recurrent kernels.
    n_gates : int
        Number of recurrent unit gates, 3 for GRU, 4 for LSTM.
    transform_func: function(numpy.ndarray)
        Function to apply to each input or recurrent kernel.

    Returns
    -------
    numpy.ndarray
        Transformed composite matrix of input or recurrent kernels in C-contiguous layout.
    """
    return np.require(np.hstack([transform_func(kernel) for kernel in np.hsplit(kernels, n_gates)]), requirements='C')


def transform_bias(bias):
    """Transforms bias weights of an LSTM layer into the regular Keras format."""
    return np.sum(np.split(bias, 2, axis=0), axis=0)


def write_text_file(path, text):
    """Write a string to a file"""
    with open(path, "w") as text_file:
        print(text, file=text_file)


def arr_as_arr5(arr):
    """Convert an n-tensor to a 5-tensor"""
    depth = len(arr.shape)
    if depth == 1:
        return arr.reshape(1, 1, 1, 1, arr.shape[0])
    if depth == 2:
        return arr.reshape(1, 1, 1, arr.shape[0], arr.shape[1])
    if depth == 3:
        return arr.reshape(1, 1, arr.shape[0], arr.shape[1], arr.shape[2])
    if depth == 4:
        return arr.reshape(1, arr.shape[0], arr.shape[1], arr.shape[2], arr.shape[3])
    if depth == 5:
        return arr
    if depth == 6 and arr.shape[0] in [None, 1]:  # todo: Is this still needed?
        return arr.reshape(arr.shape[1:])
    raise ValueError('invalid number of dimensions')


def get_layer_input_shape_shape5(layer):
    """Convert a keras shape to an fdeep shape"""
    shape = layer.input_shape[1:]
    depth = len(shape)
    if depth == 1:
        return (1, 1, 1, 1, shape[0])
    if depth == 2:
        return (1, 1, 1, shape[0], shape[1])
    if depth == 3:
        return (1, 1, shape[0], shape[1], shape[2])
    if depth == 4:
        return (1, shape[0], shape[1], shape[2], shape[3])
    if depth == 5:
        return shape
    raise ValueError('invalid number of dimensions')


def show_tensor5(tens):
    """Serialize 3-tensor to a dict"""
    values = tens.flatten()
    return {
        'shape': tens.shape,
        'values': encode_floats(values)
    }


def show_test_data_as_tensor5(arr):
    """Serialize model test data"""
    return show_tensor5(arr_as_arr5(arr))


def get_model_input_layers(model):
    """Works for different Keras version."""
    if hasattr(model, '_input_layers'):
        return model._input_layers
    if hasattr(model, 'input_layers'):
        return model.input_layers
    raise ValueError('can not get (_)input_layers from model')


def measure_predict(model, data_in):
    """Returns output and duration in seconds"""
    start_time = datetime.datetime.now()
    data_out = model.predict(data_in)
    end_time = datetime.datetime.now()
    duration = end_time - start_time
    print('Forward pass took {} s.'.format(duration.total_seconds()))
    return data_out, duration.total_seconds()


def replace_none_with(value, shape):
    """Replace every None with a fixed value."""
    return tuple(list(map(lambda x: x if x is not None else value, shape)))


def are_embedding_layer_positions_ok_for_testing(model):
    """
    Test data can only be generated if all embeddings layers
    are positioned directly behind the input nodes
    """

    def count_embedding_layers(model):
        layers = model.layers
        result = 0
        for layer in layers:
            if isinstance(layer, keras.layers.Embedding):
                result += 1
        layer_type = type(layer).__name__
        if layer_type in ['Model', 'Sequential']:
            result += count_embedding_layers(layer)
        return result

    def count_embedding_layers_at_input_nodes(model):
        result = 0
        for input_layer in get_model_input_layers(model):
            if input_layer._outbound_nodes and isinstance(
                    input_layer._outbound_nodes[0].outbound_layer, keras.layers.Embedding):
                result += 1
        return result

    return count_embedding_layers(model) == count_embedding_layers_at_input_nodes(model)


def gen_test_data(model):
    """Generate data for model verification test."""

    def set_shape_idx_0_to_1_if_none(shape):
        """Change first element in tuple to 1."""
        if shape[0] is not None:
            return shape
        shape_lst = list(shape)
        shape_lst[0] = 1
        shape = tuple(shape_lst)
        return shape

    def generate_input_data(input_layer):
        """Random data fitting the input shape of a layer."""
        if input_layer._outbound_nodes and isinstance(
                input_layer._outbound_nodes[0].outbound_layer, keras.layers.Embedding):
            random_fn = lambda size: np.random.randint(
                0, input_layer._outbound_nodes[0].outbound_layer.input_dim, size)
        else:
            random_fn = np.random.normal
        try:
            shape = input_layer.batch_input_shape
        except AttributeError:
            shape = input_layer.input_shape
        return random_fn(
            size=replace_none_with(32, set_shape_idx_0_to_1_if_none(shape))).astype(np.float32)

    assert are_embedding_layer_positions_ok_for_testing(
        model), "Test data can only be generated if embedding layers are positioned directly after input nodes."

    data_in = list(map(generate_input_data, get_model_input_layers(model)))

    warm_up_runs = 3
    test_runs = 5
    data_out = None
    for _ in range(warm_up_runs):
        measure_predict(model, data_in)
    duration_sum = 0
    print('Starting performance measurements.')
    for _ in range(test_runs):
        data_out, duration = measure_predict(model, data_in)
        duration_sum = duration_sum + duration
    duration_avg = duration_sum / test_runs
    print('Forward pass took {} s on average.'.format(duration_avg))

    return {
        'inputs': list(map(show_test_data_as_tensor5, data_in)),
        'outputs': list(map(show_test_data_as_tensor5, data_out))
    }


def split_every(size, seq):
    """Split a sequence every seq elements."""
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


def encode_floats(arr):
    """Serialize a sequence of floats."""
    if STORE_FLOATS_HUMAN_READABLE:
        return arr.flatten().tolist()
    return list(split_every(1024, base64.b64encode(arr).decode('ascii')))


def prepare_filter_weights_conv_2d(weights):
    """Change dimension order of 2d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 4
    return np.moveaxis(weights, [0, 1, 2, 3], [1, 2, 3, 0]).flatten()


def prepare_filter_weights_slice_conv_2d(weights):
    """Change dimension order of 2d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 4
    return np.moveaxis(weights, [0, 1, 2, 3], [1, 2, 0, 3]).flatten()


def prepare_filter_weights_conv_1d(weights):
    """Change dimension order of 1d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 3
    return np.moveaxis(weights, [0, 1, 2], [1, 2, 0]).flatten()


def show_conv_1d_layer(layer):
    """Serialize Conv1D layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 3
    weights_flat = prepare_filter_weights_conv_1d(weights[0])
    assert layer.padding in ['valid', 'same']
    assert len(layer.input_shape) == 3
    assert layer.input_shape[0] is None
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_conv_2d_layer(layer):
    """Serialize Conv2D layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 4
    weights_flat = prepare_filter_weights_conv_2d(weights[0])
    assert layer.padding in ['valid', 'same']
    assert len(layer.input_shape) == 4
    assert layer.input_shape[0] is None
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_separable_conv_2d_layer(layer):
    """Serialize SeparableConv2D layer to dict"""
    weights = layer.get_weights()
    assert layer.depth_multiplier == 1
    assert len(weights) == 2 or len(weights) == 3
    assert len(weights[0].shape) == 4
    assert len(weights[1].shape) == 4

    # probably incorrect for depth_multiplier > 1?
    slice_weights = prepare_filter_weights_slice_conv_2d(weights[0])
    stack_weights = prepare_filter_weights_conv_2d(weights[1])

    assert layer.padding in ['valid', 'same']
    assert len(layer.input_shape) == 4
    assert layer.input_shape[0] is None
    result = {
        'slice_weights': encode_floats(slice_weights),
        'stack_weights': encode_floats(stack_weights),
    }
    if len(weights) == 3:
        bias = weights[2]
        result['bias'] = encode_floats(bias)
    return result


def show_depthwise_conv_2d_layer(layer):
    """Serialize DepthwiseConv2D layer to dict"""
    weights = layer.get_weights()
    assert layer.depth_multiplier == 1
    assert len(weights) in [1, 2]
    assert len(weights[0].shape) == 4

    # probably incorrect for depth_multiplier > 1?
    slice_weights = prepare_filter_weights_slice_conv_2d(weights[0])

    assert layer.padding in ['valid', 'same']
    assert len(layer.input_shape) == 4
    assert layer.input_shape[0] is None
    result = {
        'slice_weights': encode_floats(slice_weights),
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_batch_normalization_layer(layer):
    """Serialize batch normalization layer to dict"""
    layer_axis = None
    if isinstance(layer.axis, int):
        layer_axis = layer.axis
    else:
        assert len(layer.axis) == 1
        layer_axis = layer.axis[0]
    assert layer_axis == -1 or layer_axis + 1 == len(layer.input_shape)
    moving_mean = K.get_value(layer.moving_mean)
    moving_variance = K.get_value(layer.moving_variance)
    result = {}
    result['moving_mean'] = encode_floats(moving_mean)
    result['moving_variance'] = encode_floats(moving_variance)
    if layer.center:
        beta = K.get_value(layer.beta)
        result['beta'] = encode_floats(beta)
    if layer.scale:
        gamma = K.get_value(layer.gamma)
        result['gamma'] = encode_floats(gamma)
    return result


def show_dense_layer(layer):
    """Serialize dense layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 2
    weights_flat = weights[0].flatten()
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_prelu_layer(layer):
    """Serialize prelu layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1
    weights_flat = weights[0].flatten()
    result = {
        'alpha': encode_floats(weights_flat)
    }
    return result


def show_embedding_layer(layer):
    """Serialize Embedding layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1
    result = {
        'weights': encode_floats(weights[0])
    }
    return result


def show_lstm_layer(layer):
    """Serialize LSTM layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 2 or len(weights) == 3
    result = {'weights': encode_floats(weights[0]),
              'recurrent_weights': encode_floats(weights[1])}

    if len(weights) == 3:
        result['bias'] = encode_floats(weights[2])

    return result


def show_gru_layer(layer):
    """Serialize GRU layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 2 or len(weights) == 3
    result = {'weights': encode_floats(weights[0]),
              'recurrent_weights': encode_floats(weights[1])}

    if len(weights) == 3:
        result['bias'] = encode_floats(weights[2])

    return result


def transform_cudnn_weights(input_weights, recurrent_weights, n_gates):
    return transform_kernels(input_weights, n_gates, transform_input_kernel), \
           transform_kernels(recurrent_weights, n_gates, transform_recurrent_kernel)


def show_cudnn_lstm_layer(layer):
    """Serialize a GPU-trained LSTM layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 3  # CuDNN LSTM always has a bias

    n_gates = 4
    input_weights, recurrent_weights = transform_cudnn_weights(weights[0], weights[1], n_gates)

    result = {'weights': encode_floats(input_weights),
              'recurrent_weights': encode_floats(recurrent_weights),
              'bias': encode_floats(transform_bias(weights[2]))}

    return result


def show_cudnn_gru_layer(layer):
    """Serialize a GPU-trained GRU layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 3  # CuDNN GRU always has a bias

    n_gates = 3
    input_weights, recurrent_weights = transform_cudnn_weights(weights[0], weights[1], n_gates)

    result = {'weights': encode_floats(input_weights),
              'recurrent_weights': encode_floats(recurrent_weights),
              'bias': encode_floats(weights[2])}

    return result


def get_transform_func(layer):
    """Returns functions that can be applied to layer weights to transform them into the standard Keras format, if applicable."""
    if layer.__class__.__name__ in ['CuDNNGRU', 'CuDNNLSTM']:
        if layer.__class__.__name__ == 'CuDNNGRU':
            n_gates = 3
        elif layer.__class__.__name__ == 'CuDNNLSTM':
            n_gates = 4

        input_transform_func = lambda kernels: transform_kernels(kernels, n_gates, transform_input_kernel)
        recurrent_transform_func = lambda kernels: transform_kernels(kernels, n_gates, transform_recurrent_kernel)
    else:
        input_transform_func = lambda kernels: kernels
        recurrent_transform_func = lambda kernels: kernels

    if layer.__class__.__name__ == 'CuDNNLSTM':
        bias_transform_func = transform_bias
    else:
        bias_transform_func = lambda bias: bias

    return input_transform_func, recurrent_transform_func, bias_transform_func


def show_bidirectional_layer(layer):
    """Serialize Bidirectional layer to dict"""
    forward_weights = layer.forward_layer.get_weights()
    assert len(forward_weights) == 2 or len(forward_weights) == 3
    forward_input_transform_func, forward_recurrent_transform_func, forward_bias_transform_func = get_transform_func(
        layer.forward_layer)

    backward_weights = layer.backward_layer.get_weights()
    assert len(backward_weights) == 2 or len(backward_weights) == 3
    backward_input_transform_func, backward_recurrent_transform_func, backward_bias_transform_func = get_transform_func(
        layer.backward_layer)

    result = {'forward_weights': encode_floats(forward_input_transform_func(forward_weights[0])),
              'forward_recurrent_weights': encode_floats(forward_recurrent_transform_func(forward_weights[1])),
              'backward_weights': encode_floats(backward_input_transform_func(backward_weights[0])),
              'backward_recurrent_weights': encode_floats(backward_recurrent_transform_func(backward_weights[1]))}

    if len(forward_weights) == 3:
        result['forward_bias'] = encode_floats(forward_bias_transform_func(forward_weights[2]))
    if len(backward_weights) == 3:
        result['backward_bias'] = encode_floats(backward_bias_transform_func(backward_weights[2]))

    return result


def get_layer_functions_dict():
    return {
        'Conv1D': show_conv_1d_layer,
        'Conv2D': show_conv_2d_layer,
        'SeparableConv2D': show_separable_conv_2d_layer,
        'DepthwiseConv2D': show_depthwise_conv_2d_layer,
        'BatchNormalization': show_batch_normalization_layer,
        'Dense': show_dense_layer,
        'PReLU': show_prelu_layer,
        'Embedding': show_embedding_layer,
        'LSTM': show_lstm_layer,
        'GRU': show_gru_layer,
        'CuDNNLSTM': show_cudnn_lstm_layer,
        'CuDNNGRU': show_cudnn_gru_layer,
        'Bidirectional': show_bidirectional_layer,
        'TimeDistributed': show_time_distributed_layer,
        'UpSampling2D': check_upsampling_2d_layer
    }


def check_upsampling_2d_layer(layer):
    print(layer.get_config())
    assert layer.get_config()['interpolation'] == 'nearest', \
        'Only interpolation nearest is currently supported by frugally-deep.'
    return None


def show_time_distributed_layer(layer):
    show_layer_functions = get_layer_functions_dict()
    config = layer.get_config()
    class_name = config['layer']['class_name']

    if class_name in show_layer_functions:

        if len(layer.input_shape) == 3:
            input_shape_new = (layer.input_shape[0], layer.input_shape[2])
        elif len(layer.input_shape) == 4:
            input_shape_new = (layer.input_shape[0], layer.input_shape[2], layer.input_shape[3])
        elif len(layer.input_shape) == 5:
            input_shape_new = (layer.input_shape[0], layer.input_shape[2], layer.input_shape[3], layer.input_shape[4])
        elif len(layer.input_shape) == 6:
            input_shape_new = (layer.input_shape[0], layer.input_shape[2], layer.input_shape[3], layer.input_shape[4],
                               layer.input_shape[5])
        else:
            raise Exception('Wrong input shape')

        layer_function = show_layer_functions[class_name]
        attributes = dir(layer.layer)

        class CopiedLayer:
            pass

        copied_layer = CopiedLayer()

        for attr in attributes:
            try:
                if attr not in ['input_shape', '__class__']:
                    setattr(copied_layer, attr, getattr(layer.layer, attr))
                elif attr == 'input_shape':
                    setattr(copied_layer, 'input_shape', input_shape_new)
            except Exception:
                continue

        setattr(copied_layer, "output_shape", getattr(layer, "output_shape"))

        return layer_function(copied_layer)

    else:
        return None


def get_dict_keys(d):
    """Return keys of a dictionary"""
    return [key for key in d]


def merge_two_disjunct_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy.
    No Key is allowed to be present in both dictionaries.
    """
    assert set(get_dict_keys(x)).isdisjoint(get_dict_keys(y))
    z = x.copy()
    z.update(y)
    return z


def is_ascii(some_string):
    """Check if a string only contains ascii characters"""
    try:
        some_string.encode('ascii')
    except UnicodeEncodeError:
        return False
    else:
        return True


def get_all_weights(model):
    """Serialize all weights of the models layers"""
    show_layer_functions = get_layer_functions_dict()
    result = {}
    layers = model.layers
    assert K.image_data_format() == 'channels_last'
    for layer in layers:
        layer_type = type(layer).__name__
        if layer_type in ['Model', 'Sequential']:
            result = merge_two_disjunct_dicts(result, get_all_weights(layer))
        else:
            if hasattr(layer, 'data_format'):
                if layer_type in ['AveragePooling1D', 'MaxPooling1D', 'AveragePooling2D', 'MaxPooling2D',
                                  'GlobalAveragePooling1D', 'GlobalMaxPooling1D', 'GlobalAveragePooling2D',
                                  'GlobalMaxPooling2D']:
                    assert layer.data_format == 'channels_last' or layer.data_format == 'channels_first'
                else:
                    assert layer.data_format == 'channels_last'

            show_func = show_layer_functions.get(layer_type, None)
            name = layer.name
            assert is_ascii(name)
            if name in result:
                raise ValueError('duplicate layer name ' + name)
            shown_layer = None
            if show_func:
                shown_layer = show_func(layer)
            if shown_layer:
                result[name] = shown_layer
            if show_func and layer_type == 'TimeDistributed':
                if name not in result:
                    result[name] = {}

                result[name]['td_input_len'] = encode_floats(np.array([len(layer.input_shape) - 1], dtype=np.float32))
                result[name]['td_output_len'] = encode_floats(np.array([len(layer.output_shape) - 1], dtype=np.float32))
    return result


def get_model_name(model):
    """Return .name or ._name or 'dummy_model_name'"""
    if hasattr(model, 'name'):
        return model.name
    if hasattr(model, '_name'):
        return model._name
    return 'dummy_model_name'


def set_model_name(model, name):
    """Overwrite .name or ._name'"""
    if hasattr(model, 'name'):
        model.name = name
    elif hasattr(model, '_name'):
        model._name = name
    else:
        pass  # Model has no name property.


def convert_sequential_to_model(model):
    """Convert a sequential model to the underlying functional format"""
    if type(model).__name__ == 'Sequential':
        name = get_model_name(model)
        if hasattr(model, '_inbound_nodes'):
            inbound_nodes = model._inbound_nodes
        elif hasattr(model, 'inbound_nodes'):
            inbound_nodes = model.inbound_nodes
        else:
            raise ValueError('can not get (_)inbound_nodes from model')
        # Since Keras 2.2.0
        if model.model == model:
            input_layer = Input(batch_shape=model.layers[0].input_shape)
            prev_layer = input_layer
            for layer in model.layers:
                prev_layer = layer(prev_layer)
            funcmodel = Model([input_layer], [prev_layer])
            model = funcmodel
        else:
            model = model.model
        set_model_name(model, name)
        if hasattr(model, '_inbound_nodes'):
            model._inbound_nodes = inbound_nodes
        elif hasattr(model, 'inbound_nodes'):
            model.inbound_nodes = inbound_nodes
    assert model.layers
    for i in range(len(model.layers)):
        if type(model.layers[i]).__name__ in ['Model', 'Sequential']:
            model.layers[i] = convert_sequential_to_model(model.layers[i])
    return model


def offset_conv2d_eval(depth, padding, x):
    """Perform a conv2d on x with a given padding"""
    kernel = K.variable(value=np.array([[[[1]] + [[0]] * (depth - 1)]]),
                        dtype='float32')
    return K.conv2d(x, kernel, strides=(3, 3), padding=padding)


def offset_sep_conv2d_eval(depth, padding, x):
    """Perform a separable conv2d on x with a given padding"""
    depthwise_kernel = K.variable(value=np.array([[[[1]] * depth]]),
                                  dtype='float32')
    pointwise_kernel = K.variable(value=np.array([[[[1]] + [[0]] * (depth - 1)]]),
                                  dtype='float32')
    return K.separable_conv2d(x, depthwise_kernel,
                              pointwise_kernel, strides=(3, 3), padding=padding)


def conv2d_offset_max_pool_eval(_, padding, x):
    """Perform a max pooling operation on x"""
    return K.pool2d(x, (1, 1), strides=(3, 3), padding=padding, pool_mode='max')


def conv2d_offset_average_pool_eval(_, padding, x):
    """Perform an average pooling operation on x"""
    return K.pool2d(x, (1, 1), strides=(3, 3), padding=padding, pool_mode='avg')


def check_operation_offset(depth, eval_f, padding):
    """Check if backend used an offset while placing the filter
    e.g. during a convolution.
    TensorFlow is inconsistent in doing so depending
    on the type of operation, the used device (CPU/GPU) and the input depth.
    """
    in_arr = np.array([[[[i] * depth for i in range(6)]]])
    input_data = K.variable(value=in_arr, dtype='float32')
    output = eval_f(depth, padding, input_data)
    result = K.eval(output).flatten().tolist()
    assert result in [[0, 3], [1, 4]]
    return result == [1, 4]


def get_shapes(tensor5s):
    """Return shapes of a list of tensors"""
    return [t['shape'] for t in tensor5s]


def calculate_hash(model):
    layers = model.layers
    hash_m = hashlib.sha256()
    for layer in layers:
        for weights in layer.get_weights():
            assert isinstance(weights, np.ndarray)
            hash_m.update(weights.tobytes())
        hash_m.update(layer.name.encode('ascii'))
    return hash_m.hexdigest()


def convert(in_path, out_path, no_tests=False):
    """Convert any Keras model to the frugally-deep model format."""

    assert K.backend() == "tensorflow"
    assert K.floatx() == "float32"
    assert K.image_data_format() == 'channels_last'

    print('loading {}'.format(in_path))
    model = load_model(in_path)

    # Force creation of underlying functional model.
    # see: https://github.com/fchollet/keras/issues/8136
    # Loss and optimizer type do not matter, since we do not train the model.
    model.compile(loss='mse', optimizer='sgd')

    model = convert_sequential_to_model(model)

    test_data = None if no_tests else gen_test_data(model)

    json_output = {}
    json_output['architecture'] = json.loads(model.to_json())

    json_output['image_data_format'] = K.image_data_format()
    for depth in range(1, 3, 1):
        json_output['conv2d_valid_offset_depth_' + str(depth)] = \
            check_operation_offset(depth, offset_conv2d_eval, 'valid')
        json_output['conv2d_same_offset_depth_' + str(depth)] = \
            check_operation_offset(depth, offset_conv2d_eval, 'same')
        json_output['separable_conv2d_valid_offset_depth_' + str(depth)] = \
            check_operation_offset(depth, offset_sep_conv2d_eval, 'valid')
        json_output['separable_conv2d_same_offset_depth_' + str(depth)] = \
            check_operation_offset(depth, offset_sep_conv2d_eval, 'same')
    json_output['max_pooling_2d_valid_offset'] = \
        check_operation_offset(1, conv2d_offset_max_pool_eval, 'valid')
    json_output['max_pooling_2d_same_offset'] = \
        check_operation_offset(1, conv2d_offset_max_pool_eval, 'same')
    json_output['average_pooling_2d_valid_offset'] = \
        check_operation_offset(1, conv2d_offset_average_pool_eval, 'valid')
    json_output['average_pooling_2d_same_offset'] = \
        check_operation_offset(1, conv2d_offset_average_pool_eval, 'same')
    json_output['input_shapes'] = list(map(get_layer_input_shape_shape5, get_model_input_layers(model)))
    if test_data:
        json_output['tests'] = [test_data]
    json_output['trainable_params'] = get_all_weights(model)
    json_output['hash'] = calculate_hash(model)

    print('writing {}'.format(out_path))
    write_text_file(out_path, json.dumps(
        json_output, allow_nan=False, indent=2, sort_keys=True))

def usage():
    print("usage: ergo to-fdeep <path>")
    quit()

def action_to_fdeep(argc, argv):
    if argc < 1:
        usage()

    prj = Project(argv[0])
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this projec")
        quit()

    convert(prj.weights_path, prj.fdeep_path)




prepare = \
"""
import pandas as pd

# this function is called whenever the `ergo train <project> --dataset file.csv`
# command is executed, the first argument is the dataset and it must return
# a pandas.DataFrame object.
def prepare_dataset(filename):
    # simply read as csv
    return pd.read_csv(filename, sep = ',', header = None)

# called during `ergo serve <project>` for each `x` input parameter, use this 
# function to convert, for instance, a file name in a vector of scalars for 
# your model input layer.
def prepare_input(x):
    # simply read as csv
    return pd.read_csv( pd.compat.StringIO(x), sep = ',', header = None)
"""

model = \
"""
import logging as log

from keras.models import Sequential
from keras.layers import Dense, Activation, Dropout

# build the model
def build_model(is_train):  
    n_inputs       = 10
    n_hidden       = (30, 30,)
    dropout        = 0.4
    activation     = 'relu'
    out_activation = 'softmax'
  
    log.info("building model for %s ..." % 'training' if is_train else 'evaluation')

    model = Sequential()
    for i, n_neurons in enumerate(n_hidden):
        # setup the input layer
        if i == 0:
            model.add(Dense(n_neurons, input_shape = (n_inputs,), activation = activation))
        else:
            model.add(Dense(n_neurons, activation = activation))
        # add dropout
        if is_train:
            model.add(Dropout(dropout))
    # setup output layer
    model.add(Dense(2, activation = out_activation))
    
    return model
"""

train = \
"""
import logging as log

from keras.callbacks import EarlyStopping

# define training strategy
def train_model(model, dataset):
    log.info("training model (train on %d samples, validate on %d) ..." % ( \\
            len(dataset.Y_train), 
            len(dataset.Y_val) ) )
    
    loss      = 'binary_crossentropy'
    optimizer = 'adam'
    metrics   = ['accuracy']
    
    model.compile(loss = loss, optimizer = optimizer, metrics = metrics)

    earlyStop = EarlyStopping(monitor = 'val_acc', min_delta=0.0001, patience = 5, mode = 'auto')
    return model.fit( dataset.X_train, dataset.Y_train,
            batch_size = 64,
            epochs = 50,
            verbose = 2,
            validation_data = (dataset.X_val, dataset.Y_val),
            callbacks = [earlyStop])
"""

Templates = { 
    'prepare.py' : prepare,
    'model.py'   : model,
    'train.py'   : train
}



import os
import logging as log
from importlib.machinery import SourceFileLoader

class Logic(object):
    @staticmethod
    def get_doer(name):
        log.debug("loading %s ..." % name)
        return SourceFileLoader("",name).load_module()

    @staticmethod
    def get_symbol(name, symbol):
        doer = Logic.get_doer(name)
        if symbol not in doer.__dict__:
            return None, "%s does not define a %s function" % (name, symbol)
        return doer.__dict__[symbol], None

    @staticmethod
    def get_symbols(name, symbols):
        doer = Logic.get_doer(name)
        ret  = []
        for symbol in symbols:
            if symbol not in doer.__dict__:
                return None, "%s does not define a %s function" % (name, symbol)
            else:
                ret.append(doer.__dict__[symbol])
        return ret, None

    def __init__(self, path):
        self.path             = os.path.abspath(path)
        self.preparer_path    = os.path.join(self.path, 'prepare.py')
        self.dataset_preparer = None
        self.input_preparer   = None
        self.builder_path     = os.path.join(self.path, 'model.py')
        self.builder          = None
        self.trainer_path     = os.path.join(self.path, 'train.py')
        self.trainer          = None

    def load(self):
        preparers, err = Logic.get_symbols(self.preparer_path, ('prepare_dataset', 'prepare_input')) 
        if err is not None:
            return err
        else:
            self.dataset_preparer, self.input_preparer = preparers

        self.builder, err = Logic.get_symbol(self.builder_path, 'build_model')
        if err is not None:
            return err

        self.trainer, err = Logic.get_symbol(self.trainer_path, 'train_model')
        if err is not None:
            return err

        return None

    def prepare_dataset(self, filename):
        return self.dataset_preparer(filename)

    def prepare_input(self, x):
        return self.input_preparer(x)








import os
import json
import argparse
import logging as log
import numpy as np

from terminaltables import AsciiTable

from ergo.project import Project

def usage():
    print("usage: ergo cmp <path_1> <path_2> --dataset <path> (--to-json <file>)")
    quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(description="comparer")
    parser.add_argument("-d", "--dataset", dest = "dataset", action = "store", type = str, required = True)
    parser.add_argument("-j", "--to-json", dest = "to_json", action = "store", type = str, required = False)
    args = parser.parse_args(argv)
    return args

def red(s):
    return "\033[31m" + s + "\033[0m"

def green(s):
    return "\033[32m" + s + "\033[0m"

# https://stackoverflow.com/questions/11942364/typeerror-integer-is-not-json-serializable-when-serializing-json-in-python
def default(o):
    if isinstance(o, np.int64): return int(o)  
    raise TypeError

def action_compare(argc, argv):
    if argc < 4:
        usage()

    args     = parse_args(argv[2:])
    metrics  = {}
    projects = { \
        argv[0]: None,
        argv[1]: None,
    }
    ref       = None
    inp_shape = None
    out_shape = None

    for path in projects:
        prj = Project(path)
        err = prj.load()
        if err is not None:
            log.error("error while loading project %s: %s", path, err)
            quit()

        if inp_shape is None:
            inp_shape = prj.model.input_shape

        elif inp_shape != prj.model.input_shape:
            log.error("model %s input shape is %s, expected %s", path, prj.model.input_shape, inp_shape)
            quit()

        if out_shape is None:
            out_shape = prj.model.output_shape

        elif out_shape != prj.model.output_shape:
            log.error("model %s output shape is %s, expected %s", path, prj.model.output_shape, out_shape)
            quit()

        if ref is None:
            ref = prj

        projects[path] = prj
        metrics[path] = None

    ref.prepare(args.dataset, 0.0, 0.0)

    log.info("evaluating %d models on %d samples ...", len(projects), len(ref.dataset.X))

    for path, prj in projects.items():
        # TODO: Run in parallel?
        log.debug("running %s ...", path)
        metrics[path] = prj.accuracy_for(ref.dataset.X, ref.dataset.Y, repo_as_dict = True)

    prev = None
    for path, m in metrics.items():
        if prev is None:
            prev = m
            continue

        ref_repo, ref_cm = prev
        new_repo, new_cm = m
        diffs = {
            'report':[],
            'cm':[],
            'cm_stats':{}
        }

        table = [ ["Name", "Ref", "New", "Delta"] ]
        for label, ref_run in ref_repo.items():
            for name, ref_value in ref_run.items():
                new_value = new_repo[label][name]
                if new_value != ref_value:
                    delta = new_value - ref_value
                    sign, fn  = ('+', green) if delta >= 0 else ('', red)
                    diffs['report'].append({
                        'name': '%s / %s' % (label, name),
                        'delta': delta,
                    })
                    table.append( [\
                        "%s / %s" % (label, name),
                        "%.2f" % ref_value,
                        "%.2f" % new_value,
                        fn("%s%.2f" % (sign, delta))] )
        print("") 
        print(AsciiTable(table).table)
        
        heads = [""]
        for i in range(0, ref_cm.shape[0]):
            heads.append("class %d" % i)

        table = [ heads ]
        total = 0
        impr  = 0
        regr  = 0
        for i in range(0, ref_cm.shape[0]):
            row = ["class %d" % i]
            row_diffs = []
            for j in range(0, ref_cm.shape[1]):
                ref_v = ref_cm[i][j]
                new_v = new_cm[i][j]
                total = total + new_v
                delta = new_v - ref_v

                if ref_v != new_v:
                    sign  = '+' if delta >= 0 else ''
                    
                    if i == j:
                        fn = green if delta >= 0 else red
                    else:
                        fn = red if delta >= 0 else green

                    if fn == green:
                        impr += abs(delta)
                    else:
                        regr += abs(delta)

                    cell = fn("%d (%s%d)" % (new_v, sign, delta))
                else:
                    cell = "%d" % ref_v

                row.append(cell)
                row_diffs.append(delta)
            
            diffs['cm'].append(row_diffs)
            table.append(row)
 
        print("")
        print(AsciiTable(table).table)

        diffs['cm_stats'] = {
            'improvements': { 
                'total': impr,
                'perc': impr / float(total) * 100.0
            },
            'regressions': {
                'total': regr,
                'perc': regr / float(total) * 100.0
            }
        }
            
        print("")
        print("Improvements: %d ( %.2f %% )" % (impr, impr / float(total) * 100.0))
        print("Regressions : %d ( %.2f %% )" % (regr, regr / float(total) * 100.0))

        if args.to_json is not None:
            print("")
            log.info("creating %s ...", args.to_json)
            with open(args.to_json, 'w+') as fp:
                json.dump(diffs, fp, default=default)



import os
import argparse

from ergo.project import Project

def usage():
    print("usage: ergo clean <path> [args]")
    quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(description="cleaner")
    parser.add_argument("-a", "--all", dest = "all", action = "store_true", default = False, help = "Remove model weights and training data.")
    args = parser.parse_args(argv)
    return args

def action_clean(argc, argv):
    if argc < 1:
        usage()

    args = parse_args(argv[1:])
    Project.clean(argv[0], args.all)



import os
import argparse
import logging as log

import pandas as pd
from flask import Flask, request, jsonify

from ergo.project import Project

prj = None
app = Flask(__name__)

@app.route('/')
def route():
    global prj

    x = request.args.get('x')
    if x is None:
        return "Missing 'x' parameter.", 400

    try:
        x = prj.logic.prepare_input(x)
        y = prj.model.predict(x)
        return jsonify(y.tolist())
    except Exception as e:
        #log.exception("error while predicting on %s", x)
        log.error("%s", e)
        return str(e), 400

def parse_args(argv):
    parser = argparse.ArgumentParser(description="server")
    parser.add_argument("--host", dest = "host", action = "store", type = str)
    parser.add_argument("--port", dest = "port", action = "store", type = int, default = 8080)
    parser.add_argument("--debug", dest = "debug", action = "store_true", default = False)
    args = parser.parse_args(argv)
    return args

def usage():
    print("usage: ergo serve <path>")
    quit()

def action_serve(argc, argv):
    global prj, app

    if argc < 1:
        usage()

    args = parse_args(argv[1:])
    prj = Project(argv[0])
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()
    elif not prj.is_trained():
        log.error("no trained Keras model found for this project")
        quit()

    app.run(host=args.host, port=args.port, debug=args.debug)



import os
import logging as log

from ergo.project import Project

def usage():
    print("usage: ergo view <path>")
    quit()

def action_view(argc, argv):
    if argc < 1:
        usage()

    prj = Project(argv[0])
    err = prj.load()
    if err is not None:
        log.error("error while loading project: %s", err)
        quit()

    prj.view()



import logging as log
import pandas as pd

def optimize_dataset(path, reuse = 0.15, output = None):
    log.info("optimizing dataset %s (reuse ratio is %.1f%%) ...", path, reuse * 100.0)

    data  = pd.read_csv(path, sep = ',', header = None)
    n_tot = len(data)

    log.info("loaded %d total samples", n_tot)

    unique  = data.drop_duplicates()
    n_uniq  = len(unique)
    n_reuse = int( n_uniq * reuse )
    reuse   = data.sample(n=n_reuse).reset_index(drop = True)

    log.info("found %d unique samples, reusing %d samples from the main dataset", n_uniq, n_reuse)

    out          = pd.concat([reuse, unique]).sample(frac=1).reset_index(drop=True)
    outpath      = output if output is not None else path
    n_out        = len(out)
    optimization = 100.0 - (n_out * 100.0) / float(n_tot)

    log.info("optimized dataset has %d records, optimization is %.2f%%", n_out, optimization)

    log.info("saving %s ...", outpath)
    out.to_csv( outpath, sep = ',', header = None, index = None)



import os
import argparse
import logging as log

from ergo.dataset import Dataset

def probability(x):
    x = float(x)
    if x < 0 or x > 1:
        raise argparse.ArgumentTypeError("%r not in range [0.0 - 1.0]" % x )
    return x

def usage():
    print("usage: ergo optimize-dataset <filename> [args]")
    quit()

def parse_args(argv):
    parser = argparse.ArgumentParser(description="optimizer")
    parser.add_argument("-r", "--reuse-ratio", dest = "reuse", action = "store", type = probability, default = 0.15)
    parser.add_argument("-o", "--output", dest = "output", action = "store", type = str, default = None)
    args = parser.parse_args(argv)
    return args

def action_optimize_dataset(argc, argv):
    if argc < 1:
        usage()

    args = parse_args(argv[1:])
    path = os.path.abspath(argv[0])
    if not os.path.exists(path):
        log.error("dataset file %s does not exist", path)
        quit()

    Dataset.optimize(path, args.reuse, args.output) 



import os
import logging as log

import numpy as np
import pandas as pd

from keras.utils import to_categorical

from ergo.core.utils import clean_if_exist
from ergo.core.optimizer import optimize_dataset
from ergo.core.saver import Saver
from ergo.core.loader import Loader

class Dataset(object):
    @staticmethod
    def clean(path):
        data_files = ( \
                'data-train.csv',
                'data-test.csv',
                'data-validation.csv',
                'data-train.pkl',
                'data-test.pkl',
                'data-validation.pkl')
        clean_if_exist(path, data_files)

    @staticmethod
    def optimize(path, reuse = 0.15, output = None):
        optimize_dataset(path, reuse, output)

    @staticmethod
    def split_row(row, n_labels, flat):
        x = row.iloc[:,1:].copy()
        if not flat:
            if len(row) == 1:
                # this check is to prevent the list comprehension to fail
                # shouldn't happen in production but can fail on test
                log.error("Dataset size must be greater than 1")
                quit()
            x = [ np.squeeze(np.array( [ x[i][:]] ), axis = 0)
                   for i in x.columns ]
        y = to_categorical(row.values[:,0], n_labels)
        return x, y

    def __init__(self, path):
        self.path       = os.path.abspath(path)
        self.train_path = os.path.join(self.path, 'data-train.csv')
        self.test_path  = os.path.join(self.path, 'data-test.csv')
        self.valid_path = os.path.join(self.path, 'data-validation.csv')
        self.saver      = Saver(self)
        self.loader     = Loader(self)
        self.do_save    = True
        self.is_flat    = True
        self.n_labels   = 0
        self.train      = None
        self.test       = None
        self.validation = None
        self.X_train    = None
        self.Y_train    = None
        self.X_test     = None
        self.Y_test     = None
        self.X_val      = None
        self.Y_val      = None
        self.X          = None
        self.Y          = None

    def exists(self):
        return os.path.exists(self.train_path) and \
               os.path.exists(self.test_path) and \
               os.path.exists(self.valid_path)

    def _set_xys(self, for_training = True):
        if for_training:
            self.X_train, self.Y_train = Dataset.split_row(self.train, self.n_labels, self.is_flat)
            self.X_test,  self.Y_test  = Dataset.split_row(self.test, self.n_labels, self.is_flat)
            self.X_val,   self.Y_val   = Dataset.split_row(self.validation, self.n_labels, self.is_flat)
        else:
            self.X, self.Y = Dataset.split_row(self.train, self.n_labels, self.is_flat)

    def load(self):
        pkl_test = self.train_path.replace('.csv', '.pkl')
        if os.path.exists(pkl_test):
            log.info("detected pickle encoded dataset")
            self.is_flat = False
            self.train_path = self.train_path.replace('.csv', '.pkl')
            self.test_path = self.test_path.replace('.csv', '.pkl')
            self.test_path = self.test_path.replace('.csv', '.pkl')

        self.loader.load()
        self._set_xys()
    
    def _is_scalar_value(self, v):
        try:
            return not (len(v) >= 0)
        except TypeError:
            # TypeError: object of type 'X' has no len()
            return True
        except:
            raise

    def source(self, data, p_test = 0.0, p_val = 0.0):
        # reset indexes and resample data just in case
        dataset = data.sample(frac = 1).reset_index(drop = True)

        # check if the input vectors are made of scalars or other vectors
        self.is_flat = True
        for x in dataset.iloc[0,:]:
            if not self._is_scalar_value(x):
                log.info("detected non scalar input: %s", x.shape)
                self.is_flat = False
                break

        # count unique labels on first column
        self.n_labels = len(dataset.iloc[:,0].unique())
        # if both values are zero, we're just loading a single file,
        # otherwise we want to generate training temporary datasets.
        for_training = p_test > 0.0 and p_val > 0.0
        if for_training:
            log.info("generating train, test and validation datasets (test=%f validation=%f) ...",
                    p_test,
                    p_val)

            n_tot   = len(dataset)
            n_train = int(n_tot * ( 1 - p_test - p_val))
            n_test  = int(n_tot * p_test)
            n_val   = int(n_tot * p_val)

            self.train      = dataset.head(n_train)
            self.test       = dataset.head(n_train + n_test).tail(n_test)
            self.validation = dataset.tail(n_val)

            if self.do_save:
                self.saver.save()
        else:
            self.train = dataset

        self._set_xys(for_training)

    def subsample(self, ratio):
        X = self.X.values if self.is_flat else self.X
        y = self.Y
        if ratio < 1.0:
            log.info("selecting a randomized sample of %d%% ...", ratio * 100)

            tot_rows = X.shape[0] if self.is_flat else X[0].shape[0]
            num      = int(tot_rows * ratio)
            indexes  = np.random.choice(tot_rows, num, replace = False)

            X = X[indexes] if self.is_flat else [ i[indexes] for i in X ]
            y = y[indexes]

        return X, y



import sys
import os
import logging as log

from ergo.project import Project

def usage():
    log.info("usage: ergo create <path>")
    quit()

def action_create(argc, argv):
    if argc != 1:
        usage()
    
    path = argv[0]

    if os.path.exists(path):
        log.error("path %s already exists" % path)
        quit()
    
    log.info("creating %s ..." % path) 
    os.makedirs(path, exist_ok=True)

    Project.create(path)







import logging as log
import pandas as pd
import numpy as np
import pickle
import threading

class Loader(object):
    def __init__(self, dataset):
        self.dataset = dataset

    def _reader(self, path):
        log.info("loading %s ..." % path)
        if self.dataset.is_flat:
            return pd.read_csv(path, sep = ',', header = None)
        else:
            return pickle.load( open( path, "rb" ) )

    def _worker(self, idx):
        if idx == 0:
            self.dataset.train = self._reader(self.dataset.train_path)
        elif idx == 1:
            self.dataset.test = self._reader(self.dataset.test_path)
        elif idx == 2:
            self.dataset.validation = self._reader(self.dataset.valid_path)

    def load(self):
        threads = ( \
          threading.Thread(target=self._worker, args=(0,)),
          threading.Thread(target=self._worker, args=(1,)),
          threading.Thread(target=self._worker, args=(2,)) 
        )
        for t in threads:
            t.start()
        for t in threads:
            t.join() 

        u = np.concatenate( \
                (self.dataset.validation.iloc[:,0].unique(), 
                self.dataset.test.iloc[:,0].unique(),
                self.dataset.train.iloc[:,0].unique()) )
        self.dataset.n_labels  = len(np.unique(u))

        log.info("detected %d distinct labels", self.dataset.n_labels)



from setuptools import setup, find_packages
from ergo.version import __version__, __author__, __email__, __license__

import os


desc = 'ergo is a tool that makes deep learning with Keras easier.'

setup( name                 = 'ergo-ai',
       version              = __version__,
       description          = desc,
       long_description     = desc,
       long_description_content_type = 'text/plain',
       author               = __author__,
       author_email         = __email__,
       url                  = 'http://www.github.com/evilsocket/ergo',
       packages             = find_packages(),
       scripts              = [ 'bin/ergo' ],
       license              = __license__,
       classifiers          = [
            'Programming Language :: Python :: 3',
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Environment :: Console',
      ],
)



import threading
import logging as log
import pickle

class Saver(object):
    def __init__(self, dataset):
        self.dataset = dataset
        self.threads = None

    @staticmethod
    def _worker(v, filename, flat):
        log.info("saving %s ..." % filename)

        if not flat:
            pickle.dump( v, open( filename, "wb" ) )
        else:
            v.to_csv(filename, sep = ',', header = None, index = None, chunksize=512)

    def save(self):
        train_path = self.dataset.train_path
        test_path  = self.dataset.test_path
        valid_path = self.dataset.valid_path
        flat = self.dataset.is_flat

        if not self.dataset.is_flat:
            train_path = train_path.replace('.csv', '.pkl')
            test_path = test_path.replace('.csv', '.pkl')
            valid_path = valid_path.replace('.csv', '.pkl')

        self.threads = [ \
          threading.Thread(target=Saver._worker, args=( self.dataset.train, train_path, flat, )),
          threading.Thread(target=Saver._worker, args=( self.dataset.test, test_path, flat, )),
          threading.Thread(target=Saver._worker, args=( self.dataset.validation, valid_path, flat, )) 
        ]

        for t in self.threads:
            t.start()
    
    def wait(self):
        if self.threads is not None:
            log.info("waiting for datasets saving to complete ...")
            for t in self.threads:
                t.join()



__author__    = 'Simone Margaritelli'
__email__     = 'evilsocket@gmail.com'
__copyright__ = 'Copyright 2019, Simone Margaritelli'
__credits__   = ['Nico Chiaraviglio', 'waxwing']
__license__   = 'GPL'
__version__   = '1.6.1'
__status__    = 'Production'

banner = """
                  ,.::XXXXXXX*:.,,,,                                                 
              .:(XXX(`        ` `*XXXX&:.,                                           
         ,.:(XX*'`                    `)*XXX::...,                                   
       *&X* `                              `#XXXXXXX*,.                              
    .X#``     .:&XX*#*:.,                       ``^:*XXXX&:.                         
  .*X`    ,:XXXXXXXXXXXXXX#.                           `^XXXX&,                      
.:X`    .XXXXXXXXXXXXXXXXXX*.                             `^XX&&&.,                  
:`     XXXXXXXXXXXXXXXXXXXXX                   .,            `XXXXX:.                
      XXXXXXXXXXXXXXXXXXXXXXX                .XXXXXX#,           `XXXXX,             
     XXXXXXXXXXXXXXXXXXXXXXXX,               .&XXX, `^X#:.          `^&XX.           
    &XXXXXXXXXXXXXXXXXX&`   `XX,             .:XXXX      `^X*.,        `^XX.         
   #XXXXXXXXXXXXXXXX&`        `XX,            `XXXX,.XXXXXXXXXXX&.,      `XX&        
   XXXXXXXXXXXXXXX(`    .u      `XXX*:.         `*X*.XXXXXXXXXXXXXXX\.    `XX&.      
  ,XXXXXXXXXXXXXX`   ud8888.        `'\*XXX&.      `*XX,XXXXXXXXXXXXXXX&,   `XX&     
  *XXXXXXXXXXXX`   :888'8888.               `XXX.     `*XX&X^`XXXXXXXXXX*"X**XXX&    
  XXXXXXXXXXX`     d888 '88*"   .u    .        `(X#,     `X&,  `XXXXXXXXX  `XXXXX&.  
  XXXXXXXX(`       8888.+"     .d88B :@8c         `XX      `X    XXXXXXXX    `XXXX&. 
 .XXXXXX:`         8888L      ="8888f8888r          `X*,     X   'XXXXXXX,      `"`  
 *XXX:`            '8888c. .+   4888>'88"      uL     `X*   &X   XXXXXXXX&,          
 XX`                "88888*     4888> '    .ue888Nc..    "*X*`   "XXX^XXX&:,         
&X`                   "YP'      4888>     d88E`"888E`       u.      `   `"`          
`                              .d888L .+  888E  888E   ...ue888b                     
                               ^"8888*"   888E  888E   888R Y888r                    
         ergo  %10s         "Y"     888E  888E   888R I888>                    
         keras %10s                 888& .888E   888R I888>                    
         tf    %10s                 *888" 888&   888R I888>                    
                                           `"   "888E  8888cJ8WX                     
                                          .dWi   `88E  "*888*P"                      
                                          4888~  J8*     'Y"                         
                                           ^"===*"`                                  
"""



import os
import shutil
import logging as log

def clean_if_exist(path, files):
    path  = os.path.abspath(path)
    for filename in files:
        filename = os.path.join(path, filename)
        if os.path.exists(filename):
            if os.path.isdir(filename):
                log.info("removing folder %s", filename)
                shutil.rmtree(filename)
            else:
                log.info("removing file %s", filename)
                os.remove(filename)
